/*
 * XLife Copyright 1989 Jon Bennett jb7m+@andrew.cmu.edu, jcrb@cs.cmu.edu
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * This module encapsulates cell-box handling. Everything else in the program
 * except possibly the evolution function should go through these functions
 * to change the board state. Entry points are:
 *
 * void initcells(context)        -- initialize active pattern
 *
 * int lookcell(context, x, y)    -- get state of cell
 *
 * void chgcell(context, x, y, c) -- set cell to state c
 *
 * void changesize(n)             -- change the cellbox allocation size
 *
 * void center()                  -- center viewport on barycenter of pattern
 *
 * void clear()                   -- clear the universe, freeing all storage
 *
 * void redisplay()               -- update visible display of active board
 *
 * void redrawscreen()            -- redraw, assuming nothing about old state
 *
 * void saveall()                 -- save entire board state to file
 *
 * This code knows that cells are packed in 8x8 tiles on a hash list, but
 * it doesn't know anything about the internals of cell representation within
 * cell boxes.  The function changesize() changes the size of cellboxes.
 *
 * This code relies on the existence of a triad of functions:
 *
 * getcell(ptr, dx, dy)      -- get state of cell at x, y in box *ptr
 * setcell(ptr, dx, dy, val) -- set cell at x, y in box *ptr to state val
 * forget(ptr)               -- cause a box to forget last generation's state
 *
 * to access cell state. It also relies on the existence of these functions:
 *
 * displaybox(x, y, ptr)     -- post box state to Xrect arrays
 * fetchtile(context, x, y)  -- get tile containing cell (x, y), NULL if none
 * maketile(context, x, y)   -- get tile holding cell (x, y), creating if none
 */

#include <stddef.h>	/* we need offsetof() */
#include <stdlib.h>
#include "defs.h"
#include "patchlevel.h"

#define MARK		'*'
#define SPACE		'.'
#define	USL_MAX		4294967295U	/* max decimal value of "unsigned" */

#ifndef NULL
#define NULL 0
#endif

/* public variables */
pattern active, tentative;
tile *freep, *curtile = NULL;
static unsigned HXM = ((1 << HASHBITS - HASHBITS/2) - 1) << 3,
   HYM = (1 << HASHBITS/2) - 1, oldxc;
static int HXS = HASHBITS/2 - 3;
static XTextProperty windowName;
static char window_name_pattern[61], *window_name = window_name_pattern,
   *zoom[] = {"8:1", "4:1", "2:1", "1:1", "1:2", "1:4", "1:8", "1:16", "1:32", "1:64"};

//#define HASH(x,y) (shl((x & HXM),HXS) + ((y >> 3) & HYM))

unsigned HASH(register unsigned x, register unsigned y) {
    return shl((x & HXM),HXS) | ((y >> 3) & HYM);
}

void hashstat(void) {
   unsigned hashsum = 0, hashcount = 0, i;
   tile *ptr;

   for (i = 0; i < HASHSIZE; i++)
      if (ptr = active.hashlist[i]) {
          do hashsum++; while (ptr = ptr->hnext);
          hashcount++;
      }
   sprintf(inpbuf, "HASH INFO: X=%u Y=%u [%u] %.4f %.3f", HASHBITS - 3 - HXS,
         3 + HXS, HASHBITS, (double)hashsum/HASHSIZE, (double)hashsum/hashcount);
   DoExpose(INPUTWIN);
}

void adjhash(pattern *p) {
   unsigned xc, yc, i, xs = HASHBITS - 3 - HXS, x,
      ys = 3 + HXS;
   tile *ptr;

   (void) bounding_box(p);
   if ((x = p->xmax - p->xmin >> 3) == 0) return;
   xc = HASHBITS/(1 + (double)(p->ymax - p->ymin >> 3)/x);
   yc = HASHBITS - xc;
   if (xc < 3) {
      xc = 3;
      yc = HASHBITS - xc;
   }
   else if (yc < 3) {
      yc = 3;
      xc = HASHBITS - yc;
   }
   if (xc == oldxc || xc >= HASHBITS || yc >= HASHBITS || p->tilecount < 2048) return;
   oldxc = xc;
   if (yc < ys && xc >= xs) {
       for (i = 0; i < ys - yc; i++) {
          HXM = (HXM << 1) | 8;
          HYM >>= 1;
          if (--HXS < -1) break;
       }
       goto adjpat;
   }
   if (xc < xs && yc >= ys) {
       for (i = 0; i < xs - xc; i++) {
          HXM = (HXM^8) >> 1;
          HYM = (HYM << 1) | 1;
          if (++HXS >= HASHBITS - xc) break;
       }
adjpat:
       memset(p->hashlist, 0, HASHSIZE*sizeof(tile*));
       for (ptr = p->tiles; ptr; ptr = ptr->next) {
          unsigned hv = HASH(ptr->x, ptr->y);

          if (ptr->hnext = p->hashlist[hv])
             ptr->hnext->hfore = ptr;
          p->hashlist[hv] = ptr;
       }
   }
}

void setscale(const int sc) {
/* preset scale for X rectangles describing active board */
    if (sc > 0) {
	int	i, j, size = (1 << sc) - (sc > 1);

	for (i = 0; i < MAXSTATES; i++)
	    for (j = 0; j < MAXCHANGE; j++) {
		rects[i][j].height = size;
		rects[i][j].width = size;
	    }
    }
    sprintf(window_name_pattern, "XLife v%d.%d.%d: a cellular-automaton laboratory [%s]",
       majorversion, minorversion, patchlevel, zoom[sc + 3]);
    if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
       (void) fprintf(stderr, "structure allocation for WindowName failed.\n");
       exit(5);
    }
    XSetWMName(disp, mainw, &windowName);
}

void initcells(pattern *context) {
/* initialize the cell hash list */
    memset(context, '\0', sizeof(pattern));
}

void chgcell(pattern *context, coord_t x, coord_t y, cell_t c) {
/* turn a cell to state ON */
    unsigned ydx, xdx;
    tile *ptr;

    ptr = maketile(context, x, y);
    ydx = y - ptr->y;
    xdx = x - ptr->x;
    if (c)
	ptr->dead = 0;

    setcell(&ptr->cells, xdx, ydx, c);
}

void changesize(int newsize) {
/* change the tile allocation size */
    if (tilesize != offsetof(tile, cells) + newsize) {
	tile	*ptr, *nptr;

	clear_pattern(&active);
	clear_pattern(&tentative);

	for (ptr = freep; ptr; ptr = nptr) {
	    nptr = ptr->next;
	    free(ptr);
	}
	freep = 0;

	tilesize = offsetof(tile, cells) + newsize;
        DoExpose(LIFEWIN);
    }
}

void killtile(pattern *context, tile *ptr) {
/* nuke a tile, reclaiming all associated storage and fixing neighbor links */
    unsigned hv = HASH(ptr->x, ptr->y);

    if (ptr != context->tiles)
	ptr->fore->next = ptr->next;
    else
	context->tiles = ptr->next;
    if (ptr == context->hashlist[hv])
	context->hashlist[hv] = ptr->hnext;
    else
	ptr->hfore->hnext = ptr->hnext;
    if (ptr->next) ptr->next->fore = ptr->fore;
    if (ptr->hnext) ptr->hnext->hfore = ptr->hfore;
    if (ptr->rt) ptr->rt->lf = NULL;
    if (ptr->lf) ptr->lf->rt = NULL;
    if (ptr->up) ptr->up->dn = NULL;
    if (ptr->dn) ptr->dn->up = NULL;
    ptr->next = freep;
    freep = ptr;
    context->tilecount--;
}

tile *createtile(pattern *context, coord_t x, coord_t y, unsigned long hv) {
/* create a tile to hold cells near life-world coordinates x, y */
/* universe it belongs in */
/* associated life-world coordinates */
/* tile-retrieval hash value */
    tile *ptr;

#ifdef PROF
    create_called++;
#endif /* PROF */

    if (freep == NULL) {
#ifdef PROF
	create_null++;
#endif /* PROF */
	if ((ptr = (tile *)malloc(tilesize)) ==  (tile *)NULL){
	    perror("create: malloc error: ");
	    exit(-1);
	}
    }
    else {
	ptr = freep;
	freep = freep->next;
    }
    memset(ptr, '\0', tilesize);

    ptr->next = context->tiles;
    context->tiles = ptr;

    if (ptr->next)
	ptr->next->fore = ptr;
    ptr->hnext = context->hashlist[hv];
    context->hashlist[hv] = ptr;
    if (ptr->hnext)
	ptr->hnext->hfore = ptr;
    ptr->x = x;
    ptr->y = y;
    context->tilecount++;
    return ptr;
}


static tile *findtile(pattern *context, coord_t x, coord_t y, unsigned hv) {
    tile *ptr = context->hashlist[hv];
#ifdef PROF
    clink_called++;
#endif /* PROF */
/*    if (freed) {  -- doesn't work with Prisoner's dilemma situation
	if (ptr==NULL){
	    ptr=freep;
	    context->hashlist[hv]= ptr;
	    freep=freep->next;
	    memset(ptr, '\0', tilesize);
	    ptr->x = x;
	    ptr->y = y;
	    ptr->next=context->tiles;
	    context->tiles=ptr;

	    if (ptr->next){
		ptr->next->fore=ptr;
	    }	
	    context->tilecount++;
	    return(ptr);
	}
	else{
	    do{
#ifdef PROF
		clink_search++;
#endif
		if ((ptr->x == x) && (ptr->y == y)){
		    return(ptr);
		}
		ptr=ptr->hnext;
	    }while (ptr!=NULL);
	    
	    return(NULL);
	}
    }
    else*/
    {
	while (ptr) {
#ifdef PROF
	    clink_search++;
#endif /* PROF */
	    if (ptr->x == x && ptr->y == y)
		return ptr;
	    ptr = ptr->hnext;
	}
	
	return 0;
    }
}

int lookcell(pattern *context, coord_t x, coord_t y) {
/* get the state of a cell from its world coordinates */
    tile *ptr;
    unsigned hv;
    coord_t hx = x & 0xfffffff8, hy = y & 0xfffffff8;

    hv = HASH(x, y);
    if (ptr = findtile(context, hx, hy, hv))
	return getcell(&ptr->cells, x - hx, y - hy);
    else
	return 0;
}

tile *fetchtile(pattern *context, coord_t x, coord_t y) {
    return findtile(context, x, y, HASH(x, y));
}

tile *maketile(pattern *context, coord_t x, coord_t y) {
    tile *ptr;
    unsigned hv;

    hv = HASH(x, y);
    x &= 0xfffffff8;
    y &= 0xfffffff8;
    if (ptr = findtile(context, x, y, hv));
    else
       ptr = createtile(context, x, y, hv);
    ptr->act = 1;
    return ptr;
}

void center(void) {
/* center the display on the `center of mass' of the live boxes */
    double x = 0, y = 0;
    int ctr = 0;
    tile *ptr;

    for (ptr = active.tiles; ptr; ptr = ptr->next)
	if (!ptr->dead){
	    x += ptr->x;
	    y += ptr->y;
	    ctr++;
	}
    if (ctr > 0) {
       x /= ctr;
       y /= ctr;
    }
    else
      x = xorigin, y = yorigin;
    xpos = x - shr((unsigned)width/2, scale);
    ypos = y - shr((unsigned)height/2, scale);
    DoExpose(LIFEWIN);
}

int cmpnum(const void *e1, const void *e2) {
  if (*(coord_t*)e1 < *(coord_t*)e2)
    return -1;
  return *(coord_t*)e1 > *(coord_t*)e2;
}

void median() {
/* return the median coordinates of the live cells in the context */
    coord_t *coordxlist, *coordylist;
    int ctr = 0, x = xorigin, y = yorigin;
    tile *ptr;

    if (!(coordxlist = (coord_t*)malloc(sizeof(coord_t)*active.tilecount)))
	return;
    if (!(coordylist = (coord_t*)malloc(sizeof(coord_t)*active.tilecount))) {
        free(coordxlist);
	return;
    }
    for (ptr = active.tiles; ptr != NULL; ptr = ptr->next){
	if(!ptr->dead) {
	    coordxlist[ctr] = ptr->x;
	    coordylist[ctr++] = ptr->y;
	}
    }
    if (ctr > 0) {
	qsort(coordxlist, ctr, sizeof(coord_t), cmpnum);
	qsort(coordylist, ctr, sizeof(coord_t), cmpnum);
	x = coordxlist[ctr/2];
	y = coordylist[ctr/2];
    }
    free(coordylist);
    free(coordxlist);

    xpos = x - shr((unsigned)width/2, scale);
    ypos = y - shr((unsigned)height/2, scale);
    DoExpose(LIFEWIN);
}

void current_tile() {
    if (curtile == NULL)
       curtile = active.tiles;
    if (curtile != NULL) {
       coord_t xnew, ynew;
       int lflag = 0;
       xnew = curtile->x - shr((unsigned)width/2, scale);
       ynew = curtile->y - shr((unsigned)height/2, scale);
       while (curtile->dead || labs((long)xnew - (long)xpos) < 10 << 6 - scale && labs((long)ynew - (long)ypos) < 10 << 6 - scale) {
          curtile = curtile->next;
          if (curtile == NULL) {
             if (lflag) return;
             curtile = active.tiles;
             lflag++;
          }
          xnew = curtile->x - shr((unsigned)width/2, scale);
          ynew = curtile->y - shr((unsigned)height/2, scale);
       }
       xpos = xnew;
       ypos = ynew;
       curtile = curtile->next;
       DoExpose(LIFEWIN);
    }
}

cellcount_t cellcnt(pattern *pp) {
   int i = 1;
   cellcount_t sum = 0;

   for (; i < maxstates; i++)
      sum += active.cellcount[i];
   return sum;
}

void displaystats(void) {
    sprintf(inpbuf, "Generations: %lu, Tiles: %lu", active.generations, active.tilecount);
    if (dispboxes)
	sprintf(inpbuf + strlen(inpbuf), ", Cells: %lu", cellcnt(&active));
    if (dispchanges)
        sprintf(inpbuf + strlen(inpbuf), ", Changes: %lu", active.chgcount);
    DoExpose(INPUTWIN);
}

void clear_pattern(pattern *pp) {
/* free given pattern */
    tile *ptr, *nptr;

    initcells(pp);
    ptr = pp->tiles;
    while (ptr) {
	nptr = ptr->next;
	free(ptr);
	ptr = nptr;
    }
    pp->tiles = NULL;

    pp->generations = 0;
    pp->tilecount = 0;
    memset(pp->cellcount, 0, sizeof(cellcount_t)*maxstates);
}

void clear(void) {
/* clear the board or tentative pattern, freeing all storage */
    if (tentative.tiles) {
	clear_pattern(&tentative);
	free_loadscript();    		/* restart load script */

	XClearWindow(disp, lifew);
	XClearWindow(disp, inputw);
    }
    else
    {
	clear_pattern(&active);
	state = STOP;

	XClearWindow(disp, lifew);
	XClearWindow(disp, inputw);
	displaystats();
    }
    showgrid = 1;
    redisplay();
}

#define ONSCREEN(x, y)	(((x > xpos-BOXSIZE) && (x < (xpos+BOXSIZE+shr((unsigned)width, scale))))\
   && ((y > ypos-BOXSIZE) && (y < (ypos+BOXSIZE+shr((unsigned)height,scale)))))

#define POINTS	4000	/* accumulate this many point changes before writing */
#define RECTS	64	/* accumulate this many box changes before writing */

#define SETCC(s) if (s == 1 && maxstates == 2 && multicolor) \
                           cc = cellgc[active.generations%15 + 1]; \
                        else cc = cellgc[s]

void redisplay(void) {
/* re-display the visible part of the board */
    tile *ptr;
    coord_t x, y;
    int ctr = 0, curstate;
    long tentx, tenty;
    GC cc;

    displaystats();
    if (state == HIDE) return;

    if (dispmesh && scale > 1 && showgrid) {
       coord_t cx = width >> scale, cy = height >> scale;
       int gcind, small = 1 << scale;

       showgrid = 0;
       for (x = 0; x <= cx; x++) {
            gcind = x%5 ? GRID_COLOR : GRID_ECOLOR;

            /*
             * This line is doubled in order to cope with the nearly 2:1
             * aspect ratio of typical pixels.
             */
            /*XDrawLine(disp, lifew, cellgc[gcind], x*small - 2, 0, x*small - 2, height);*/
            XDrawLine(disp, lifew, cellgc[gcind], x*small - 1, 0, x*small - 1, height);
        }
        for (y = 0; y <= cy; y++) {
            gcind = y%5 ? GRID_COLOR : GRID_ECOLOR;
            XDrawLine(disp, lifew, cellgc[gcind], 0, y*small - 1, width, y*small - 1);
        }
    }

    if (tentative.tiles && !wireframe)
        boxpattern(1); /* white outline box */

#if STATEBITS == 1
    chgpoints[0] = chgpoints[1] = chgrects[0] = chgrects[1] = 0;
#else
    for (curstate = 0; curstate < maxstates; curstate++)
	chgpoints[curstate] = chgrects[curstate] = 0;
#endif

    for (ptr = active.tiles; ptr; ptr = ptr->next) {
	x = ptr->x;
	y = ptr->y;
	if (ONSCREEN(x, y)) {
	    displaybox(x, y, &ptr->cells);
	    ctr++;
	    if (scale <= 0) {
#if STATEBITS == 1
		if (chgpoints[1] >= POINTS || chgpoints[0] >= POINTS) {
		    XDrawPoints(disp, lifew, cellgc[0], points[0], chgpoints[0], CoordModeOrigin);
		    XDrawPoints(disp, lifew, cellgc[1], points[1], chgpoints[1], CoordModeOrigin);
		    chgpoints[1] = chgpoints[0] = 0;
		}
#else
		for (curstate = 0; curstate < maxstates; curstate++)
		    if (chgpoints[curstate]) {
                        SETCC(curstate);
			XDrawPoints(disp, lifew, cc,
				    points[curstate], chgpoints[curstate], CoordModeOrigin);
			chgpoints[curstate] = 0;
		    }
#endif
		ctr = 0;
	    }
	    else if (ctr == RECTS) {
#if STATEBITS == 1
		    XFillRectangles(disp, lifew, cellgc[0], rects[0], chgrects[0]);
		    XFillRectangles(disp, lifew, cellgc[1], rects[1], chgrects[1]);
		    chgrects[1] = chgrects[0] = 0;
#else
		    for (curstate = 0; curstate < maxstates; curstate++) {
                        SETCC(curstate);
			XFillRectangles(disp, lifew, cc,
					rects[curstate], chgrects[curstate]);
			chgrects[curstate] = 0;
		    }
#endif
		    ctr = 0;
		}
	}
    }
    if (ctr) {
         for (curstate = 0; curstate < maxstates; curstate++) {
            SETCC(curstate);
            XFillRectangles(disp, lifew, cc,
                                rects[curstate], chgrects[curstate]);
            chgrects[curstate] = 0;
         }
         ctr = 0;
    }
    /* draw tentative points with appropriate transformation */
    for (ptr = tentative.tiles; ptr; ptr = ptr->next) {
	tentx = ptr->x - STARTX;
	tenty = ptr->y - STARTY;
	if (ONSCREEN(tx(tentx,tenty,txx,txy) + loadx,
		     ty(tentx,tenty,tyx,tyy) + loady)) {
	    trdisplaybox(tentx, tenty, &ptr->cells);
	    ctr++;
	    if (scale <= 0) {
#if STATEBITS == 1
		if ((chgpoints[1] >= POINTS) || (chgpoints[0] >= POINTS)) {
		    XDrawPoints(disp, lifew, cellgc[0], points[0], chgpoints[0], CoordModeOrigin);
		    XDrawPoints(disp, lifew, cellgc[1], points[1], chgpoints[1], CoordModeOrigin);
		    chgpoints[1] = chgpoints[0] = 0;
		}
#else
		for (curstate = 0; curstate < maxstates; curstate++)
		    if (chgpoints[curstate]) {
			XDrawPoints(disp, lifew, cellgc[curstate],
				    points[curstate], chgpoints[curstate],
				    CoordModeOrigin);
			chgpoints[curstate] = 0;
		    }
#endif
		ctr = 0;
	    }
	    else if (ctr == RECTS) {
#if STATEBITS == 1
		    XFillRectangles(disp, lifew, cellgc[0], rects[0], chgrects[0]);
		    XFillRectangles(disp, lifew, cellgc[1], rects[1], chgrects[1]);
		    chgrects[1] = chgrects[0] = 0;
#else
		    for (curstate = 0; curstate < maxstates; curstate++) {
                        if (wireframe)
                           XDrawRectangles(disp, lifew, cellgc[LOAD_BOX],
                                        rects[curstate], chgrects[curstate]);
                        else
			   XFillRectangles(disp, lifew, cellgc[curstate],
					rects[curstate], chgrects[curstate]);
			chgrects[curstate] = 0;
		    }
#endif
		    ctr = 0;
		}
	}
    }
    if (ctr) {
	 for (curstate = 0; curstate < maxstates; curstate++) {
            if (wireframe)
                  XDrawRectangles(disp, lifew, cellgc[LOAD_BOX],
                                        rects[curstate], chgrects[curstate]);
            else
	          XFillRectangles(disp, lifew, cellgc[curstate],
				rects[curstate], chgrects[curstate]);
	    chgrects[curstate] = 0;
	 }
    }
    /* draw pivot of tentative points and box them */
    if (tentative.tiles)
	drawpivot(shl(loadx - xpos, scale), shl(loady - ypos, scale));
    XFlush(disp);
}

void redrawscreen(void) {
/* force redraw of entire board */
    tile *ptr;

    for (ptr = active.tiles; ptr; ptr = ptr->next)
	forget(&ptr->cells);
    for (ptr = tentative.tiles; ptr; ptr = ptr->next)
	forget(&ptr->cells);
    redisplay();
#if STATEBITS > 1
    DoExpose(INPUTWIN);		/* refresh the color picker */
#endif				/* STATEBITS > 1 */
}

cellcount_t bounding_box(pattern *context) {
/* get the bounding box of a pattern */
    int dx, dy, state;
    tile *ptr;

    context->xmin = USL_MAX;
    context->ymin = USL_MAX;
    context->xmax = 0;
    context->ymax = 0; 
    for (state = 1; state < maxstates; state++)
       context->cellcount[state] = 0;
    FOR_CELLS(context, ptr, dx, dy)
	if (state = getcell(&ptr->cells, dx, dy)) {
	    context->cellcount[state]++;
	    if (ptr->x + dx < context->xmin)
	        context->xmin = ptr->x + dx;
	    if (ptr->y + dy < context->ymin)
		context->ymin = ptr->y + dy;
	    if (ptr->x + dx > context->xmax)
	        context->xmax = ptr->x + dx;
	    if (ptr->y + dy > context->ymax)
	        context->ymax = ptr->y + dy;
	}

    return cellcnt(context);
}

static void flood_fill(pattern *from, coord_t x, coord_t y, pattern *to) {
/* transplant all points in a blob touching (x,y) between contexts */
    static offsets neighbors[] =
    {
	{-1, -1},
	{-1,  0},
	{-1,  1},
	{ 0, -1},
	{ 0,  1},
	{ 1, -1},
	{ 1,  0},
	{ 1,  1},
#ifdef DEPTH2
	{-2, -2},
	{-2, -1},
	{-2,  0},
	{-2,  1},
	{-2,  2},
	{-1,  2},
	{ 0,  2},
	{ 1,  2},
	{ 2,  2},
	{ 2,  1},
	{ 2,  0},
	{ 2, -1},
	{ 2, -2},
	{ 1, -2},
	{ 0, -2},
	{-1, -2},
#endif /* DEPTH2 */
    };
    int i;

    /* skip empty cells */
    if (!lookcell(from, x, y))
	return;

    /* skip cells that are already colored */
    if (lookcell(to, x, y))
	return;

    /* transfer this cell from the copy to the new context */
    chgcell(to,   x, y, lookcell(from, x, y));
    chgcell(from, x, y, 0);

    for (i = 0; i < sizeof(neighbors)/sizeof(neighbors[0]); i++) {
	int nx = x + neighbors[i].xi;
	int ny = y + neighbors[i].yi;

	/* recursively transfer all its neighbors */
	flood_fill(from, nx, ny, to);
    }
}

static int refcmp(const void *r1, const void *r2) {
/* sort blobs first by least Y coordinate, then by least X coordinate */
    int	cmp = (((pattern*)r1)->ymin - ((pattern*)r2)->ymin);

    if (!cmp)
	cmp = (((pattern*)r1)->xmin - ((pattern*)r2)->xmin);
    return cmp;
}

pattern *analyze(pattern *context) {
/* analyze a pattern into an allocated list of blobs */
    cell_t	state;
    tile	*ptr;
    int		dx, dy, color = 0;
    pattern	analysand;
    pattern	*parts, *pp;

    initcells(&analysand);

    /* copy the pattern */
    FOR_CELLS(context, ptr, dx, dy)
	if ((state = getcell(&ptr->cells, dx, dy)))
	    chgcell(&analysand, ptr->x + dx, ptr->y + dy, state);

    /*
     * This loop is deceptively simple-looking.  It relies on flood_fill()
     * not returning until it has transpanted the largest blob it can find
     * connected to a cell.
     */
    parts = (pattern *)calloc(sizeof(pattern), 1);
    FOR_CELLS(&analysand, ptr, dx, dy)
	if (getcell(&ptr->cells, dx, dy)) {
	    parts = (pattern *)realloc(parts, sizeof(pattern) * (color + 1));

	    initcells(&parts[color]);
	    flood_fill(&analysand, ptr->x + dx, ptr->y + dy, &parts[color]);

	    color++;
	}

    /* create an extra, blank pattern entry as a fencepost */
    parts = (pattern *)realloc(parts, sizeof(pattern) * (color + 1));
    initcells(&parts[color]);

    /* this cleans up the blob order -- strictly cosmetic */
    for (pp = parts; pp->tiles; pp++)
	(void) bounding_box(pp);
    qsort(parts, color, sizeof(pattern), refcmp);

    /* return the analyzed parts */
    return parts;
}

void savesimple(pattern *pp, int xoff, int yoff, FILE *ofp) {
/* save a pattern as a simple picture block */
    coord_t x, y;
    cellcount_t cellcount = bounding_box(pp);

    if (cellcount == 0)
	return;

    if (!xoff && !yoff)
	fputs("#P\n", ofp);
    else
	fprintf(ofp, "#P %d %d\n", xoff, yoff);
    for (y = pp->ymin; y <= pp->ymax; y++) {
	for (x = pp->xmin; x <= pp->xmax; x++) {
	    int val;

	    if (val = lookcell(pp, x, y)) {
#if STATEBITS > 1
		if (maxstates > 2)
		    (void) fputc(itos(val), ofp);
		else
#endif
		    (void) fputc(MARK, ofp);
	    }
	    else
		(void) fputc(SPACE, ofp);
	}
	(void) fputc('\n', ofp);
    }
}

static void savecomplex(pattern *context, FILE *ofp) {
/* do a pattern analysis and save the pattern as a blob sequence */
    pattern	*pp, *parts = analyze(context);
    int		xcenter, ycenter, blobcount = 0;

    if (parts == NULL) {
	error("Pattern analysis failed; too many blobs in picture.\n");
	return;
    }

    xcenter = context->xmax/2 + context->xmin/2;
    ycenter = context->ymax/2 + context->ymin/2;

    for (pp = parts; pp->tiles; pp++)
	blobcount++;
    if (blobcount > 1)
	fprintf(ofp, "#C %d blobs\n", blobcount);
    for (pp = parts; pp->tiles; pp++)
	savesimple(pp, pp->xmin - xcenter, pp->ymin - ycenter, ofp);
}

int rlechr(int c) {
   if (c == 0)
      return 'b';
   else if (c == 1)
      return 'o';
   return c + '@';
}

void saveall(FILE *ofp, char mode) {
/* save the entire board contents (or the tentative pattern if there is one) */
    tile *ptr;
    char buf[80] = "", *p = buf;
    int dx, dy, val;
    coord_t x = 0, y;
    cellcount_t cellcount;
    pattern	*pp = tentative.tiles ? &tentative : &active;

    if (fname[0])
        fprintf(ofp,"#N %s\n", fname);
    else if (stashed[0]) {
        char *p = strrchr(stashed, '/');
        if (p)
             p++;
        else
             p = stashed;
        if (strchr(p, ':'))
             fprintf(ofp,"#N %s\n", p);
    }
    stamp("#0", ofp);
    if (outcome[0])
        fprintf(ofp,"#K%s\n", outcome);
    while (x < numcomments)
	fprintf(ofp, "#C %s\n", comments[x++]);
    if (fname[0] != 0)
	fprintf(ofp, "#N %s\n", fname);

#if STATEBITS > 1
    if (maxstates > 2 && active_rules[0]) {
        char c = 'U';
        if (strchr(active_rules, '$') || active_rules == strstr(active_rules, "wireworld")
               || active_rules == strstr(active_rules, "brian's brain") || strchr(active_rules, '/'))
           c = 'T';
	(void) fprintf(ofp, "#%c %s\n", c, active_rules);
    }
    else if (active_rules[0])
#endif				/* STATEBITS > 1 */
        (void) fprintf(ofp, "#T %s\n", active_rules);
    /*
     * Determine the bounding box of the live cells.
     */
    cellcount = bounding_box(pp);

    /* caller may want save to shortest format */
    if (mode == '\0')
	if ((pp->ymax - pp->ymin + 1) * (pp->xmax - pp->xmin + 1) > cellcount * 8)
	    mode = 'R';
	else
	    mode = 'P';

    /* now that we have the bounding box, we can gen the other formats */
    switch (mode) {
	case 'A': /* save in absolute mode */
	  (void) fputs("#A\n", ofp);
	  FOR_CELLS(pp, ptr, dx, dy)
			if (val = getcell(&ptr->cells, dx, dy))
#if STATEBITS > 1
			    if (maxstates > 2)
				(void) fprintf(ofp, "%d %d %d\n",
					       ptr->x + dx - xorigin, 
                                               ptr->y + dy - yorigin, val);
			    else
#endif
				(void) fprintf(ofp,
					       "%d %d\n", 
                                               ptr->x + dx - xorigin, 
                                               ptr->y + dy - yorigin);
	return;

	case 'R':
	  (void) fprintf(ofp, "#R\n");
	  FOR_CELLS(pp, ptr, dx, dy)
			if (val = getcell(&ptr->cells, dx, dy))
#if STATEBITS > 1
			    if (maxstates > 2)
				(void) fprintf(ofp, "%d %d %d\n",
					       ptr->x + dx - pp->xmin, ptr->y + dy - pp->ymin,
					       val);
			    else 
#endif
				(void) fprintf(ofp, "%d %d\n",
					       ptr->x + dx - pp->xmin, ptr->y + dy - pp->ymin);
	  return;

	case 'P':
	  savesimple(pp, 0, 0, ofp);
	  return;

	case 'D':
	  (void) fprintf(ofp, "#D\n");
	  x = pp->xmin;
	  y = pp->ymin;
	  FOR_CELLS(pp, ptr, dx, dy)
			if (val = getcell(&ptr->cells, dx, dy)) {
#if STATEBITS > 1
			    if (maxstates > 2)
				(void) fprintf(ofp, "%d %d %d\n",
					       ptr->x + dx - x, ptr->y + dy - y,
					       val);
			    else
#endif
				(void) fprintf(ofp, "%d %d\n",
					       ptr->x + dx - x, ptr->y + dy - y);
			    x = ptr->x + dx;
			    y = ptr->y + dy;
			}
	  return;

	case 'M':
	  (void) fprintf(ofp, "#M\n");
	  dy = 0;
	  for (y = pp->ymin; y <= pp->ymax; y++) {
	    for (x = pp->xmin; x <= pp->xmax; x = dx) {
		val = lookcell(pp, x, y);
		for (dx = x + 1; dx <= pp->xmax && lookcell(pp, dx, y) == val; dx++);
		if (val) {
		    if (dy > 1)
			sprintf(p, "%d$", dy);
		    else if (dy > 0)
			sprintf(p, "$");
		    p = buf + strlen(buf);
		    dy = 0;
#if STATEBITS > 1
		    if (maxstates > 2)
			if (dx - x > 1)
			    (void) sprintf(p, "%d%c", dx - x, rlechr(val));
			else
			    (void) sprintf(p, "%c", rlechr(val));
		    else
#endif
			if (dx - x > 1)
			    (void) sprintf(p, "%do", dx - x);
			else
			    (void) sprintf(p, "o");
		}
		else if (dx <= pp->xmax) {
		    if (dy > 1)
			sprintf(p, "%d$", dy);
		    else if (dy > 0)
			sprintf(p, "$");
		    p = buf + strlen(buf);
		    dy = 0;
		    if (dx - x > 1)
		        (void) sprintf(p, "%db", dx - x);
		    else
		        (void) sprintf(p, "b");
		}
	        p = buf + strlen(buf);
		if (strlen(buf) > 65) {
		    fprintf(ofp, "%s\n", p = buf);
		    buf[0] = 0;
		}
	    }
	    dy++;
	  }
	  if (buf[0])
	    fprintf(ofp, "%s", buf);
	  fprintf(ofp, "!\n");
	  return;

	case 'I':
	  savestructured(pp, ofp);
	  return;

	case 'S':
	  savecomplex(pp, ofp);
	  return;
    }
}

/* tile.c ends here */
